rm(list = ls())
library(MASS)
library(nor1mix)
library(kedd)

"kde"<-function(xin, xout, h)
{
  n<- length(xin)
  arg1<-(sapply(xout, "-", xin))/h  
  arg2<-dnorm(arg1)
  arg3<-colSums(arg2)
  arg3/(n*h)
}

hopt<-function(xin,  ML.est,   sig.lev )
{
  n<-length(xin)
  h<-bw.nrd(xin) #bw.SJ(xin)
  DensityEst <- kde(xin,xin, h )
  Rhatfx <- mean(DensityEst)
  RhatfxUse<-Rhatfx^{3/2}
  RK<- 1/(2*sqrt(pi))  # 3/5 for epanechnikov,  1/(2*sqrt(pi)) #for gaussian, 1/2 # for uniform,
  RKuse<-RK^{3/2} 
  ConvK<- kernel.conv(kernel.conv(0, deriv.order = 0, kernel =  "gaussian")$kx, deriv.order = 0, kernel =  "gaussian")$kx #1/(2*sqrt(2*pi)) #fConvolution(kernel)
  
  #theta.hat<- norMixMLE(xin, m, trace=-1) 
  #ML.est <- norMix(name = "ML.est",  mu = theta.hat[,1],  sigma = theta.hat[,2],   w = theta.hat[,3])
  
  ML.Dens <-dnorMix(xin,  ML.est) #p1 ML.est # dnorMix(xin,  p1) #p1 ML.est
  
   
  mu.2 <- mean((DensityEst - ML.Dens)^2)
  nu.2 <- mean(DensityEst^2)
  
  SigmaSq <- 2* mu.2^2 *  nu.2 * RK # 16/45 * RK * Rhatfx
  term1 <- ( (sqrt(2)*ConvK*nu.2)/(3*RKuse*RhatfxUse) )^{-1/2} #
  RDeltaR<- mean( (ML.Dens - DensityEst)^2 * DensityEst )  # \Delta_n term  mean( (dnorMix(xin, MW.nm2) - DensityEst)^2 * DensityEst )
  term2<- ( 1/(    sqrt(2*   nu.2 * RK)) )^{-3/2} 
  banduse1 <- (n*RDeltaR)^{-3/2} *  6* ConvK^{-1/2} * RK^{3/2} * nu.2 * Rhatfx^{3/4}  * SigmaSq^{3/2} #return simplified  bandwidth: for Epanechnikov kernel,  
  banduse<-   term1 * (n*RDeltaR)^{-3/2} *term2 # *term2 #return    bandwidth: for Epanechnikov kernel,
  banduse
}

T.n2<-function(xin, ML.est, h )
  #xin: the random sample
  # h: the bandwidth to use in the test statistic - also conider the very first MSE optimal rule
  # p1: 1st parameter of the distribution dist. In case of Normal Mixtures this is the object MW.nm7 etc
{
  h1<-bw.SJ(xin)
  yi <- kde(xin,xin, h1)
  #theta.hat<- norMixMLE(xin, m, trace=-1) 
  #ML.est <- norMix(name = "ML.est",  mu = theta.hat[,1],  sigma = theta.hat[,2],   w = theta.hat[,3])
  
  Null.Dens.Est <-dnorMix(xin,  ML.est) #p1 ML.est # dnorMix(xin,  p1) #p1 ML.est
  Dens.Diffs<- yi-Null.Dens.Est  # Yi - f(x_i) 
  All.Dens.Diffs<- Dens.Diffs %*% t(Dens.Diffs)  # cross product of all Yi - f(x_i) diffs
  arg1<- sapply(xin, "-", xin)  # cross product of all BinCenters in order to calculate K{(x_i - x_j)/h}
  arg2<-dnorm(arg1/h)
  test.stat.tmp <- arg2 * All.Dens.Diffs
  m2hat <- mean(yi - Null.Dens.Est)
  test.statistic <- sum(test.stat.tmp)/(length(xin) * m2hat* sqrt(h))  
  test.statistic 
}

Boot.cutoff<-function(SampleSize , M, ML.est,    a.sig)
{
  x3<-rnorMix(SampleSize, ML.est)
  h.use.edg2 <- hopt(x3,   ML.est, a.sig)
  resamples <- lapply(1:M, function(k) x3[sample.int(SampleSize, replace=T)] )
  Tn.Boot<-sapply(resamples, T.n2, ML.est=ML.est, h=h.use.edg2)
  quantile( Tn.Boot, probs=  1-a.sig , type=7)
}



gal <- galaxies/1000  
c(width.SJ(gal, method = "dpi"), width.SJ(gal))
plot(x = c(0, 40), y = c(0, 0.3), type = "n", bty = "l", xlab = "velocity of galaxy (1000km/s)", ylab = "density", ylim=c(0, 0.25))
rug(gal)
lines(density(gal, width = width.SJ(gal), n = 200), lty = 1)
#lines(density(gal, width = 2.56, n = 200), lty = 3)

m<- 6
theta.hat<- norMixMLE(gal, m, trace=-1) 
ML.est <- norMix(name = "ML.est",  mu = theta.hat[,1],  sigma = theta.hat[,2],   w = theta.hat[,3])
lines(ML.est,  p.norm = FALSE, lty=2, lwd=2, col=2)
h.use.edg <- hopt(gal,   ML.est,    a.sig)
Test.Statistic<- T.n2(gal, ML.est, h.use.edg )
Band.Cutoff<- Boot.cutoff(length(gal) , 1000, ML.est,    0.95)
 
cat("Test Statistic = ", Test.Statistic, " Band.Cutoff = ", Band.Cutoff,"\n")
 
 
m<- 5
theta.hat<- norMixMLE(gal, m, trace=-1) 
ML.est <- norMix(name = "ML.est",  mu = theta.hat[,1],  sigma = theta.hat[,2],   w = theta.hat[,3])
lines(ML.est,   p.norm = FALSE, lty=3, lwd=1.2, col=3)

m<- 4
theta.hat<- norMixMLE(gal, m, trace=-1) 
ML.est <- norMix(name = "ML.est",  mu = theta.hat[,1],  sigma = theta.hat[,2],   w = theta.hat[,3])
lines(ML.est,   p.norm = FALSE, lty=4,  lwd=1.2, col=4)
legend("topright", legend = c("kde", "MLE m=6", "MLE m=5", "MLE m=4"), col = 1:4, lty=1:4, lwd=c(1,2,1.2,1.2), ncol=2)
